/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include "nxpcanvas.h"
#include <gdk/gdkkeysyms.h>
#include "nxpcanvascontainer.h"
#include "nxplot-artft.h"

enum {
    PROP_0,
    PROP_PLOTTER
};


#define NXP_CANVAS_OVERLAY_PRIORITY G_PRIORITY_LOW
#define NXP_CANVAS_UPDATE_PRIORITY (GTK_PRIORITY_RESIZE + 1)


/*  static guint signals[LAST_SIGNAL] = { 0 }; */
static NxpCanvasClass *nxp_canvas_class = NULL;
static gpointer parent_class = NULL;


static void     nxp_canvas_init                 (NxpCanvas        *instance,
						 gpointer          g_class);
static void     nxp_canvas_class_base_init      (NxpCanvasClass   *class);
static void     nxp_canvas_class_init           (NxpCanvasClass   *class,
						 gpointer          data);
static void     nxp_canvas_shutdown             (GObject          *object);
static gint     button_press_event              (GtkWidget        *widget,
						 GdkEventButton   *event);
static gint     button_release_event            (GtkWidget        *widget,
						 GdkEventButton   *event);
static gint     motion_notify_event             (GtkWidget        *widget,
						 GdkEventMotion   *event);
static gint     expose_event                    (GtkWidget        *widget,
						 GdkEventExpose   *event);
static void     nxp_canvas_leave                (NxpCanvas        *self,
						 GdkEvent         *event);
static void     nxp_canvas_enter                (NxpCanvas        *self,
						 GdkEvent         *event);
/* static void     nxp_canvas_leave_object         (NxpCanvas        *self, */
/* 						 NxpCanvasObject  *obj); */
/* static void     nxp_canvas_enter_object         (NxpCanvas        *self, */
/* 						 NxpCanvasObject  *obj); */
static void     nxp_canvas_draw_object          (NxpCanvas        *self,
						 NxpCanvasObject  *obj,
						 Nxplot           *pl);
static gboolean idle_redraw_overlays            (NxpCanvas        *self);
void            nxp_canvas_queue_draw_overlay   (NxpCanvas        *self,
						 NxpCanvasOverlay *overlay,
						 gboolean          generate_expose);
static gboolean key_press_event                 (NxpCanvas        *self,
						 GdkEventKey      *event);
static void     set_property                    (GObject          *object,
						 guint             prop_id,
						 const GValue     *value,
						 GParamSpec       *pspec);
static void     get_property                    (GObject          *object,
						 guint             prop_id,
						 GValue           *value,
						 GParamSpec       *pspec);
static gboolean draw_idle_func                  (NxpCanvas        *self);






static GTypeInfo typeinfo = 
{
    sizeof(NxpCanvasClass), 
    (GBaseInitFunc)nxp_canvas_class_base_init, 
    NULL,
    (GClassInitFunc)nxp_canvas_class_init, 
    NULL,
    NULL,
    sizeof(NxpCanvas),
    0,
    (GInstanceInitFunc)nxp_canvas_init,
    NULL,
};



GType nxp_canvas_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(GTK_TYPE_DRAWING_AREA, "NxpCanvas",
				      &typeinfo, 0);
    }
    return type;
}


static void nxp_canvas_realize(GtkWidget *widget)
{
    NxpCanvas *self = NXP_CANVAS(widget);

    if (GTK_WIDGET_CLASS(parent_class)->realize)
	GTK_WIDGET_CLASS(parent_class)->realize(widget);
    self->xor_gc = gdk_gc_new(widget->window);
    gdk_gc_copy(self->xor_gc, widget->style->white_gc);
    gdk_gc_set_function(self->xor_gc, GDK_XOR);
}


static void nxp_canvas_init(NxpCanvas *self, gpointer g_class)
{
    if (!nxp_canvas_class)
	nxp_canvas_class = NXP_CANVAS_GET_CLASS(self);
    self->plotter = NULL;
    self->view.x1 = 0; self->view.y1 = 0;
    self->view.x1 = 1; self->view.y1 = 1;
    self->update_sigid = 0;
    self->objects = NULL;
    self->drawable_objects = NULL;

    gtk_signal_connect(GTK_OBJECT(self), "expose_event",
		       (GtkSignalFunc) expose_event, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "motion_notify_event",
		       (GtkSignalFunc) motion_notify_event, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "button_press_event",
		       (GtkSignalFunc) button_press_event, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "button_release_event",
		       (GtkSignalFunc) button_release_event, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "key_press_event",
		       (GtkSignalFunc) key_press_event, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "leave_notify_event",
		       (GtkSignalFunc) nxp_canvas_leave, NULL);
    gtk_signal_connect(GTK_OBJECT(self), "enter_notify_event",
		       (GtkSignalFunc) nxp_canvas_enter, NULL);
    gtk_widget_set_events(GTK_WIDGET(self), GDK_EXPOSURE_MASK
			  | GDK_LEAVE_NOTIFY_MASK
			  | GDK_ENTER_NOTIFY_MASK
			  | GDK_BUTTON_PRESS_MASK
			  | GDK_BUTTON_RELEASE_MASK
			  | GDK_POINTER_MOTION_MASK
			  | GDK_POINTER_MOTION_HINT_MASK
			  | GDK_KEY_PRESS_MASK);
    GTK_WIDGET_SET_FLAGS(GTK_WIDGET(self), GTK_CAN_FOCUS);
    GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(self), GTK_DOUBLE_BUFFERED);

    self->root = NULL;
    self->overlay_idle_id = 0;
    self->update_idle_id = 0;
}

static void nxp_canvas_class_base_init(NxpCanvasClass *class)
{
    /* empty */
}


static void nxp_canvas_class_init(NxpCanvasClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS(class);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(class);

    parent_class = g_type_class_peek_parent (class);
    object_class->set_property = set_property;
    object_class->get_property = get_property;

    object_class->finalize = nxp_canvas_shutdown;
    widget_class->realize = nxp_canvas_realize;

    g_object_class_install_property(object_class,
				    PROP_PLOTTER,
				    g_param_spec_object(
					"plotter", "Plotter",
					"Plotter object for screen drawing",
					TYPE_NXPLOT,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));

}


static void nxp_canvas_shutdown(GObject *object)
{
    NxpCanvas *self;
    GSList *l;
    self = NXP_CANVAS(object);
    /* disconnect signals */
    if (self->update_sigid) {
	g_signal_handler_disconnect(self->plotter, self->update_sigid);
	self->update_sigid = 0;
    }
    /* remove overlay drawing idle function */
    if (self->overlay_idle_id)
	gtk_idle_remove(self->overlay_idle_id);
    self->overlay_idle_id = 0;

    /* remove update idle function */
    if (self->update_idle_id)
	gtk_idle_remove(self->update_idle_id);
    self->update_idle_id = 0;

    /* delete plotter */
    if (self->plotter)
	g_object_unref(self->plotter);
    self->plotter = NULL;
    /* unref overlays and free list */
    for (l = self->overlays; l; l = l->next)
	g_object_unref(G_OBJECT(l->data));
    g_slist_free(self->overlays);
    self->overlays = NULL;
    /* unref objects and free list */
    for (l = self->objects; l; l = l->next)
	g_object_unref(G_OBJECT(l->data));
    g_slist_free(self->objects);
    self->objects = NULL;
    if (G_OBJECT_CLASS(parent_class)->finalize)
	G_OBJECT_CLASS(parent_class)->finalize(object);
}


GtkWidget* nxp_canvas_new(void)
{
    GTypeInstance *instance;
    NxpCanvas *self;

    instance = g_type_create_instance(nxp_canvas_get_type());
    self = NXP_CANVAS(instance);

    return GTK_WIDGET(self);
}


static void nxp_canvas_update(Nxplot *pl, NxpCanvas *self)
{
    /* GSList *l; */
    /* NxpCanvasObject *object; */
    /* NxpCanvasOverlay *overlay; */

    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(IS_NXPLOT(pl));
    g_return_if_fail(pl == self->plotter);

    if (g_type_is_a(G_OBJECT_TYPE(pl), TYPE_NXPLOT_ARTFT)) {
	if (!nxplot_artft_ready(NXPLOT_ARTFT(pl)))
	    return;
    }

    /* remove update idle function */
    if (self->update_idle_id)
	gtk_idle_remove(self->update_idle_id);
    self->update_idle_id = 0;

    /* Make sure all overlays are hidden, carefully so that do not try
       to hide themselves (they would only end up drawing
       themselves). */
/*     for (l = self->overlays; l; l = l->next) { */
/* 	overlay = NXP_CANVAS_OVERLAY(l->data); */
/* 	overlay->visible = FALSE; */
/*     } */
/*     g_slist_free(self->overlays); */
/*     self->overlays = NULL; */

    if (self->root)
	nxp_canvas_object_configure(NXP_CANVAS_OBJECT(self->root),
				    &self->view, NXP_COBJ_INTERACTION_MOVE_RESIZE);
    nxplot_open(pl);
    nxplot_erase(pl);
/*     for (l = self->objects; l; l = l->next) { */
/* 	object = NXP_CANVAS_OBJECT(l->data); */
/* 	nxp_canvas_object_update(object); */
/*     } */
    nxplot_close(pl);
    gtk_widget_queue_draw(GTK_WIDGET(self));
}


static void set_plotter(NxpCanvas *self, Nxplot *plotter)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(plotter == NULL || IS_NXPLOT(plotter));

    if (self->update_sigid) {
	g_signal_handler_disconnect(plotter, self->update_sigid);
    }
    if (self->plotter)
	g_object_unref(self->plotter);
    self->plotter = plotter;
    if (self->plotter) {
        g_object_ref(plotter);
        self->update_sigid = g_signal_connect_data(plotter, "update",
                                                   (GCallback)nxp_canvas_update,
                                                   self, NULL, 0);
    }
}

/**
 * nxp_canvas_set_plotter: -set the renderer object
 * @self: the canvas
 * @plotter: a #Nxplot renderer that must already be attached to @self
 * 
 * Sets the plotter that will be used for rendering the canvas (for
 * display, *not* printing).
 **/
void nxp_canvas_set_plotter(NxpCanvas *self, Nxplot *plotter)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(plotter);
    g_return_if_fail(IS_NXPLOT(plotter));

    set_plotter(self, plotter);
    g_object_notify(G_OBJECT(self), "plotter");
}


/**
 * nxp_canvas_draw_to:
 * @self: the canvas
 * @plotter: a #NxpPlotter to render with/into.
 * 
 * Renders the canvas with @plotter. This method could be used, for
 * example, to print the canvas into a postscript libplot plotter, or
 * anything else one can imagine.  The destination rectangle is the
 * same as the canvas' view coordinates.
 **/
void nxp_canvas_draw_to(NxpCanvas *self, Nxplot *pl)
{
    Nxplot *pl_save;
    GSList *l;
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(IS_NXPLOT(pl));

    pl_save = self->plotter;
    self->plotter = pl;
    
    if (self->root)
	nxp_canvas_object_configure(NXP_CANVAS_OBJECT(self->root),
				    &self->view, NXP_COBJ_INTERACTION_MOVE_RESIZE);
    for (l = self->drawable_objects; l; l = l->next)
	nxp_canvas_draw_object(self, NXP_CANVAS_OBJECT(l->data), pl);
    self->plotter = pl_save;
    nxp_canvas_queue_update(self);
}

/* Returns list with container followed by children;
   each container child is recusively expanded. */
static GSList *_expand_container(NxpCanvasContainer *container)
{
    GSList *children = nxp_canvas_container_get_children(container);
    GSList *l;
    GSList *result   = NULL;
    NxpCanvasObject *child;

    if (!NXP_CANVAS_OBJECT(container)->visible) return NULL;

    for (l = children; l; l = l->next) {
	child = NXP_CANVAS_OBJECT(l->data);
	if (g_type_is_a(G_OBJECT_TYPE(child), NXP_TYPE_CANVAS_CONTAINER))
	    result = g_slist_concat(_expand_container(NXP_CANVAS_CONTAINER(child)), result);
	else {
	    if (!NXP_CANVAS_OBJECT(child)->visible) continue;
	    if (NXP_CANVAS_OBJECT(child)->drawable)
		result = g_slist_prepend(result, child);
	}
    }

    if (NXP_CANVAS_OBJECT(container)->drawable)
	result = g_slist_prepend(result, container);

    g_slist_free(children);
    return result;
}


void nxp_canvas_recompute_drawable_objects_list(NxpCanvas *self)
{
    g_return_if_fail(NXP_IS_CANVAS(self));

    if (self->drawable_objects) {
	g_slist_free(self->drawable_objects);
	self->drawable_objects = NULL;
    }
    if (NXP_CANVAS_OBJECT(self->root)->visible)
	self->drawable_objects = _expand_container(self->root);
/*     g_message("Canvas %p has %i visible objects", self, */
/* 	      g_slist_length(self->drawable_objects)); */
}


static gboolean _nxp_canvas_update_idle_func(NxpCanvas *self)
{
    g_return_val_if_fail(NXP_IS_CANVAS(self), FALSE);
    self->update_idle_id = 0;
    if (self->plotter)
	nxp_canvas_update(self->plotter, self);
    return FALSE;
}


void nxp_canvas_queue_update(NxpCanvas *self)
{
    if (!self->update_idle_id)
	self->update_idle_id =
	    gtk_idle_add_priority(NXP_CANVAS_UPDATE_PRIORITY,
				  (GtkFunction) _nxp_canvas_update_idle_func,
				  self);
}


void nxp_canvas_add_object(NxpCanvas *self, NxpCanvasObject *object)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(object));
    g_return_if_fail(!object->canvas || object->canvas == self);
    g_object_ref(object);
    if (object->canvas != self)
	nxp_canvas_object_set_canvas(object, self);
    if (!g_slist_find(self->objects, object))
	self->objects = g_slist_append(self->objects, object);
    nxp_canvas_recompute_drawable_objects_list(self);
    nxp_canvas_queue_update(self);
}


void nxp_canvas_remove_object(NxpCanvas *self, NxpCanvasObject *object)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(object));
    g_return_if_fail(object->canvas == self);

    self->objects = g_slist_remove(self->objects, object);
    object->canvas = NULL;
    g_object_unref(object);
    nxp_canvas_recompute_drawable_objects_list(self);
    nxp_canvas_queue_update(self);
}

static gint button_press_event(GtkWidget *widget, GdkEventButton *event)
{
    NxpCanvas *self = NXP_CANVAS(widget);

/*     g_message(__FUNCTION__); */
    if (!GTK_WIDGET_HAS_FOCUS(widget)) {
/* 	g_message("Setting focus."); */
	gtk_widget_grab_focus(GTK_WIDGET(widget));
    }
    nxp_canvas_object_button_press(NXP_CANVAS_OBJECT(self->root), (GdkEvent *) event);
    return TRUE;
}


static gint button_release_event(GtkWidget *widget, GdkEventButton *event)
{
    NxpCanvas *self = NXP_CANVAS(widget);
    nxp_canvas_object_button_release(NXP_CANVAS_OBJECT(self->root), (GdkEvent *) event);
    return TRUE;
}


__inline__ gboolean gdk_rectangle_point_in(const GdkRectangle *rect, gint x, gint y)
{
    gint w = x - rect->x, h = y - rect->y;
    return (w >= 0 && h >= 0 && w <= rect->width && h <= rect->height);
}


void nxp_canvas_get_pointer(NxpCanvas *self, gint *x, gint *y, GdkModifierType *state)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(GTK_WIDGET(self)->window);
    gdk_window_get_pointer(GTK_WIDGET(self)->window, x, y, state);
}


static gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
    NxpCanvas *self = NXP_CANVAS(widget);
    gint x, y;
    if (event->is_hint) {
	gdk_window_get_pointer(event->window, &x, &y, &event->state);
	event->x = x;
	event->y = y;
    }
    nxp_canvas_object_motion_notify(NXP_CANVAS_OBJECT(self->root), (GdkEvent *) event);
    return FALSE;
}



/**
 * expose_event:
 * @widget: 
 * @event: 
 * 
 * Monitor expose events, but not handle them, only to determine if
 * any overlay object will be overwritten by such an expose. Any such
 * overlay needs to be queued for redraw.
 * 
 * Return value: 
 **/
static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    NxpCanvas *self = NXP_CANVAS(widget);
    GSList *l;
    NxpCanvasOverlay *overlay;
    GdkRectangle rect;
    gboolean completely_inside;

/*     g_message("Expose area: (%i,%i,%i,%i)", */
/*  	      event->area.x, event->area.y, */
/*  	      event->area.width, event->area.height); */
/*     if (!self->overlays) g_message("Overlay list empty!"); */
    for (l = self->overlays; l; l = l->next) {
	overlay = NXP_CANVAS_OVERLAY(l->data);
	if (gdk_rectangle_intersect(&overlay->bbox, &event->area, &rect)) {
/*  	    g_message("Expose area (%i,%i,%i,%i) intercepted object %p," */
/*  		      " which has bbox = (%i,%i,%i,%i)", */
/*  		      event->area.x, event->area.y, */
/*  		      event->area.width, event->area.height, */
/*  		      overlay, overlay->bbox.x, overlay->bbox.y, */
/*  		      overlay->bbox.width, overlay->bbox.height); */
	    /* Check if the current expose event completely covers the
               overlay */
	    completely_inside = 
		overlay->bbox.x >= event->area.x &&
		overlay->bbox.y >= event->area.y &&
		overlay->bbox.x + overlay->bbox.width <=
		event->area.x + event->area.width &&
		overlay->bbox.y + overlay->bbox.height <=
		event->area.y + event->area.height;
	    /* If it only covers the overlay partially, an extra
               expose event needs to be generated to delete the
               overlay completely. */
	    nxp_canvas_queue_draw_overlay(self, overlay, !completely_inside);
	} else {
/*  	    g_message("Expose area (%i,%i,%i,%i) did NOT intercept object %p," */
/*  		      " which has bbox = (%i,%i,%i,%i)", */
/*  		      event->area.x, event->area.y, */
/*  		      event->area.width, event->area.height, */
/*  		      overlay, overlay->bbox.x, overlay->bbox.y, */
/*  		      overlay->bbox.width, overlay->bbox.height); */
	}
    }
    /* Make sure our RGB buffer is up-to-date before handling expose
       events. */
    draw_idle_func(self);
    return FALSE;
}


static void nxp_canvas_enter(NxpCanvas *self, GdkEvent *event)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
/*     g_message(__FUNCTION__); */
    nxp_canvas_object_enter(NXP_CANVAS_OBJECT(self->root));
}


static void nxp_canvas_leave(NxpCanvas *self, GdkEvent *event)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
/*     g_message(__FUNCTION__); */
    nxp_canvas_object_leave(NXP_CANVAS_OBJECT(self->root));
}


static void nxp_canvas_draw_object(NxpCanvas *self, NxpCanvasObject *obj, Nxplot *pl)
{
    nxplot_savestate(pl);
    nxp_canvas_object_draw(obj, pl);
    nxplot_restorestate(pl);
}



static void nxp_canvas_redraw_area1(NxpCanvas *self, GdkRectangle *area)
{
    GSList *l;
    NxpCanvasObject *obj;
    GdkRectangle tmp;

    /* Get a list of all objects that need redrawing, and those that
       recursively will need redraw too (for intersecting objects),
       and mark them all as 'dirty'. */
    for (l = self->drawable_objects; l; l = l->next) {
	obj = NXP_CANVAS_OBJECT(l->data);
	if (obj->dirty) {
	    continue;
	}
	if (gdk_rectangle_intersect(area, &obj->wbbox, &tmp)) {
	    obj->dirty = TRUE;
	    nxp_canvas_redraw_area1(self, &obj->wbbox);
	} 
    }
}



void nxp_canvas_redraw_area(NxpCanvas *self, NxplotRect *area)
{
    GSList *l;
    NxpCanvasObject *obj;
    GdkRectangle warea;

    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));

    if (g_type_is_a(G_OBJECT_TYPE(self->plotter), TYPE_NXPLOT_ARTFT)) {
	if (!nxplot_artft_ready(NXPLOT_ARTFT(self->plotter)))
	    return;
    }

    nxplot_rect_space2window(self->plotter, area, &warea);
    /* Mark the objects that need redraw. */
    nxp_canvas_redraw_area1(self, &warea);
    /* Clear all the areas beneath the dirty objects plus the original
       requested area */
    nxplot_open(self->plotter);
    for (l = self->drawable_objects; l; l = l->next) {
	obj = NXP_CANVAS_OBJECT(l->data);
	if (obj->dirty)
	    nxplot_clearrect(self->plotter, obj->bbox.x1, obj->bbox.y1,
			      obj->bbox.x2, obj->bbox.y2);
    }
    nxplot_clearrect(self->plotter, area->x1, area->y1,
		      area->x2, area->y2);
    nxplot_close(self->plotter);
}


void nxp_canvas_add_overlay(NxpCanvas *self, NxpCanvasOverlay *overlay)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(NXP_IS_CANVAS_OVERLAY(overlay));
    g_object_ref(overlay);
/*     g_message(__FUNCTION__); */
    self->overlays = g_slist_append(self->overlays, overlay);
    overlay->dirty = FALSE;
    /* draw to screen */
    nxp_canvas_draw_overlay(self, overlay);
}


void nxp_canvas_remove_overlay(NxpCanvas *self, NxpCanvasOverlay *overlay)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(NXP_IS_CANVAS_OVERLAY(overlay));

/*     g_message(__FUNCTION__); */

    g_return_if_fail(g_slist_find(self->overlays, overlay));

    self->overlays = g_slist_remove(self->overlays, overlay);
    nxp_canvas_draw_overlay(self, overlay);
/*     g_message("%s: unref'ing overlay with refcount=%i", */
/* 	      __FUNCTION__, G_OBJECT(overlay)->ref_count); */
    g_object_unref(overlay);
}


void nxp_canvas_draw_overlay(NxpCanvas *self, NxpCanvasOverlay *overlay)
{
    g_return_if_fail(self->xor_gc);
    nxp_canvas_overlay_draw(overlay, GTK_WIDGET(self)->window, self->xor_gc, 0);
}


void nxp_canvas_queue_draw_overlay(NxpCanvas *self, NxpCanvasOverlay *overlay,
				   gboolean generate_expose)
{
    GSList *l;
    NxpCanvasOverlay *overlay1;
    GdkRectangle rect;
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(NXP_IS_CANVAS_OVERLAY(overlay));

    if (overlay->dirty)
	return;
    overlay->dirty = TRUE;
    if (!self->overlay_idle_id)
	self->overlay_idle_id = gtk_idle_add_priority(
	    NXP_CANVAS_OVERLAY_PRIORITY, (GtkFunction)idle_redraw_overlays, self);
    if (generate_expose) {
	gtk_widget_queue_draw_area(GTK_WIDGET(self), overlay->bbox.x, overlay->bbox.y,
				   overlay->bbox.width, overlay->bbox.height);
    }
    /* Check to see what other overlays are affected (intersect) by
       this area. They will also need redraw. */
    for (l = self->overlays; l; l = l->next) {
	overlay1 = NXP_CANVAS_OVERLAY(l->data);
	/* Infinite recursion should not happen because of the 'dirty'
           flag on each overlay. */
	if (gdk_rectangle_intersect(&overlay->bbox, &overlay1->bbox, &rect))
	    nxp_canvas_queue_draw_overlay(self, overlay1, TRUE);
    }
}


/**
 * idle_redraw_overlays:
 * @self: 
 * 
 * Draw all overlays that have the 'dirty' flag set. This function
 * runs as an idle handler with lower priority (i.e. after) than normal
 * drawing.
 **/
static gboolean idle_redraw_overlays(NxpCanvas *self)
{
    GSList *l;
    NxpCanvasOverlay *overlay;
    g_return_val_if_fail(NXP_IS_CANVAS(self), FALSE);
    self->overlay_idle_id = 0;
    
    for (l = self->overlays; l; l = l->next) {
	overlay = NXP_CANVAS_OVERLAY(l->data);
	if (overlay->dirty && overlay->visible) {
	    nxp_canvas_draw_overlay(self, overlay);
	    overlay->dirty = FALSE;
	}
    }
    return FALSE;
}


void nxp_canvas_set_cursor(NxpCanvas *self, GdkCursor *cur)
{
    g_return_if_fail(NXP_IS_CANVAS(self));
    if (cur != self->current_cursor) {
	self->current_cursor = cur;
	if (GTK_WIDGET_REALIZED(self))
	    gdk_window_set_cursor(GTK_WIDGET(self)->window, cur);
    }
}


static gboolean draw_idle_func(NxpCanvas *self)
{
    GSList *l;
    NxpCanvasObject *obj;

    if (g_type_is_a(G_OBJECT_TYPE(self->plotter), TYPE_NXPLOT_ARTFT)) {
	if (!nxplot_artft_ready(NXPLOT_ARTFT(self->plotter)))
	    return FALSE;
    }

    nxplot_open(self->plotter);
    for (l = self->drawable_objects; l; l = l->next) {
	obj = NXP_CANVAS_OBJECT(l->data);
	if (obj->dirty) {
	    nxp_canvas_draw_object(self, obj, self->plotter);
	    obj->dirty = FALSE;
	    gtk_widget_queue_draw_area(GTK_WIDGET(self), obj->wbbox.x,
				       obj->wbbox.y, obj->wbbox.width,
				       obj->wbbox.height);
	}
    }
    nxplot_close(self->plotter);
    return FALSE;
}



static gboolean key_press_event(NxpCanvas *self, GdkEventKey *event)
{
    if (event->keyval == GDK_b && event->state&GDK_CONTROL_MASK) {
	GSList *l;
	for (l = self->objects; l; l = l->next)
	    nxp_canvas_object_enter(NXP_CANVAS_OBJECT(l->data));
	return TRUE;
    }
    if (self->root)
	nxp_canvas_object_key_press(NXP_CANVAS_OBJECT(self->root), (GdkEvent *) event);
    return TRUE;
}



static void
set_property (GObject      *object,
	      guint         prop_id,
	      const GValue *value,
	      GParamSpec   *pspec)
{
    NxpCanvas *self;

    self = NXP_CANVAS(object);

    switch (prop_id)
    {

    case PROP_PLOTTER:
	set_plotter(self, NXPLOT(g_value_get_object(value)));
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}



static void
get_property (GObject      *object,
                           guint         prop_id,
                           GValue       *value,
                           GParamSpec   *pspec)
{
    NxpCanvas *self = NXP_CANVAS(object);

    switch (prop_id)
    {
    case PROP_PLOTTER:
	if (self->plotter)
	    g_value_set_object(value, G_OBJECT(self->plotter));
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}


/**
 * nxp_canvas_set_root:
 * @self: a #NxpCanvas
 * @root_container: #a NxpCanvasContainer; the canvas takes ownership of this object reference
 * 
 * Sets @root_container as the root container for the canvas
 * @self. This container will thereon be accessible as self->root.
 * This function should be called only once, i.e. you can't change the
 * root container.
 **/
void nxp_canvas_set_root(NxpCanvas *self,  NxpCanvasContainer *root_container)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(root_container);
    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(root_container));
    g_return_if_fail(!self->root);

    self->root = root_container;
    if (NXP_CANVAS_OBJECT(root_container)->canvas) {
	g_assert(NXP_CANVAS_OBJECT(root_container)->canvas == self);
    } else
	nxp_canvas_add_object(self, NXP_CANVAS_OBJECT(self->root));
}


void nxp_canvas_set_viewport(NxpCanvas *self, double x1, double y1,
			     double x2, double y2)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS(self));
    g_return_if_fail(x2 > x1);
    g_return_if_fail(y2 > y1);
    self->view.x1 = x1;
    self->view.x2 = x2;
    self->view.y1 = y1;
    self->view.y2 = y2;
    gtk_widget_queue_resize(GTK_WIDGET(self));
}


