/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include "nxpcanvascontainer.h"
#include "nxplot-marshalers.h"
#include "nxpcanvasmisc.h"

enum {
    GET_CHILDREN_SIGNAL,
    GET_CHILD_AT_POINT_SIGNAL,    
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 }; 

enum {
    PROP_0,
};


static gpointer parent_class = NULL;



static void     nxp_canvas_container_init            (NxpCanvasContainer      *instance,
						      gpointer                 g_class);
static void     nxp_canvas_container_class_base_init (NxpCanvasContainerClass *class);
static void     nxp_canvas_container_class_init      (NxpCanvasContainerClass *class,
						      gpointer                 data);
static void     nxp_canvas_container_shutdown        (GObject                 *object);
static void     set_property                         (GObject                 *object,
						      guint                    prop_id,
						      const GValue            *value,
						      GParamSpec              *pspec);
static void     get_property                         (GObject                 *object,
						      guint                    prop_id,
						      GValue                  *value,
						      GParamSpec              *pspec);
static void     nxp_canvas_container_enter           (NxpCanvasObject         *obj);
static void     nxp_canvas_container_leave           (NxpCanvasObject         *obj);
static gboolean nxp_canvas_container_button_press    (NxpCanvasObject         *obj,
						      GdkEvent                *event);
static gboolean nxp_canvas_container_button_release  (NxpCanvasObject         *obj,
						      GdkEvent                *event);
static gboolean nxp_canvas_container_motion_notify   (NxpCanvasObject         *obj,
						      GdkEvent                *event);
static gboolean nxp_canvas_container_key_press       (NxpCanvasObject         *obj,
						      GdkEvent                *event);
static void     nxp_canvas_container_update          (NxpCanvasObject         *obj);



static NxpCanvasObject* 
nxp_canvas_container_real_get_child_at_point(NxpCanvasContainer *self,
					     gint x, gint y);




static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxpCanvasContainerClass), 

    (GBaseInitFunc)nxp_canvas_container_class_base_init, 
    NULL, // GBaseFinalizeFunc      base_finalize;

    /* classed types, instantiated types */
    (GClassInitFunc)nxp_canvas_container_class_init, 
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;

    /* instantiated types */
    sizeof(NxpCanvasContainer),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxp_canvas_container_init, //GInstanceInitFunc      instance_init;

    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};


GType nxp_canvas_container_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(NXP_TYPE_CANVAS_OBJECT, "NxpCanvasContainer",
				      &typeinfo, 0);
    }
    return type;
}


static void nxp_canvas_container_init(NxpCanvasContainer *self, gpointer g_class)
{
/*     g_message(__FUNCTION__); */

}


static void nxp_canvas_container_class_base_init(NxpCanvasContainerClass *class)
{
    /* empty */
}


static void nxp_canvas_container_class_init(NxpCanvasContainerClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS(class);
    NxpCanvasObjectClass *cobj_class = NXP_CANVAS_OBJECT_CLASS(class);

    object_class->set_property = set_property;
    object_class->get_property = get_property;
    object_class->finalize = nxp_canvas_container_shutdown;
    
    cobj_class->enter = nxp_canvas_container_enter;
    cobj_class->leave = nxp_canvas_container_leave;
    cobj_class->button_press = nxp_canvas_container_button_press;
    cobj_class->button_release = nxp_canvas_container_button_release;
    cobj_class->motion_notify = nxp_canvas_container_motion_notify;
    cobj_class->key_press = nxp_canvas_container_key_press;
    cobj_class->update = nxp_canvas_container_update;

    class->get_child_at_point = nxp_canvas_container_real_get_child_at_point;

    parent_class = g_type_class_peek_parent (class);

    signals[GET_CHILDREN_SIGNAL] =
	g_signal_new("get-children",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasContainerClass, get_children),
		     NULL, NULL,
		     _nxplot_marshal_BOXED__VOID,
		     TYPE_NXPLOT_LIST, 0);
    signals[GET_CHILD_AT_POINT_SIGNAL] =
	g_signal_new("get-child-at-point",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasContainerClass, get_child_at_point),
		     _nxp_object_handled_accumulator, NULL,
		     _nxplot_marshal_OBJECT__INT_INT,
		     NXP_TYPE_CANVAS_OBJECT, 2, G_TYPE_INT, G_TYPE_INT);


}


static void nxp_canvas_container_shutdown(GObject *object)
{
/*     NxpCanvasContainer *self; */
/*     self = NXP_CANVAS_CONTAINER(object); */

    if (G_OBJECT_CLASS(parent_class)->finalize)
	G_OBJECT_CLASS(parent_class)->finalize(object);
}


static void
set_property (GObject      *object,
	      guint         prop_id,
	      const GValue *value,
	      GParamSpec   *pspec)
{
    NxpCanvasContainer *self;

    self = NXP_CANVAS_CONTAINER(object);

    switch (prop_id)
    {


    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}



static void
get_property (GObject      *object,
                           guint         prop_id,
                           GValue       *value,
                           GParamSpec   *pspec)
{
    NxpCanvasContainer *self;

    self = NXP_CANVAS_CONTAINER(object);

    switch (prop_id)
    {

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}


/**
 * nxp_canvas_container_get_focus_object:
 * @self: a container
 * 
 * Get object that has focus, i.e. the object that should receive all keyboard events.
 * 
 * Return value: the focused object
 **/
__inline__ static NxpCanvasObject*
nxp_canvas_container_get_focus_object(NxpCanvasContainer *self)
{
    if (self->grab_object)
	return self->grab_object;
    else
	return self->in_object;
}


static void nxp_canvas_container_enter(NxpCanvasObject *obj)
{
    gint x, y;
    GdkModifierType mask;
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(obj);
    nxp_canvas_get_pointer(NXP_CANVAS(obj->canvas), &x, &y, &mask);
    nxp_canvas_container_check_mouse_over(self, x, y);
}


static void nxp_canvas_container_leave(NxpCanvasObject *obj)
{
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(obj);
    if (self->in_object) {
	nxp_canvas_object_leave(self->in_object);
	self->in_object = NULL;
    }
}


/* This is a default implementation that subclasses can override in
 * order to provide a more efficient/customized implementation. */
static NxpCanvasObject* 
nxp_canvas_container_real_get_child_at_point(NxpCanvasContainer *self,
					      gint x, gint y)
{
    NxpCanvasObject *rv = NULL;
    GSList *l;
    g_return_val_if_fail(self, NULL);
    g_return_val_if_fail(NXP_IS_CANVAS_CONTAINER(self), NULL);
    l = nxp_canvas_container_get_children(self);
    if (!l) return NULL;
    for (; l; l = l->next)
	if (gdk_rectangle_point_in(&NXP_CANVAS_OBJECT(l->data)->wbbox, x, y)) {
	    rv = NXP_CANVAS_OBJECT(l->data);
	    break;
	}
    g_slist_free(l);
    return rv;
}


/**
 * nxp_canvas_container_get_child_at_point:
 * @self: the container
 * @x: x
 * @y: y
 * 
 * Find out which child object, if any, is at the given coordinates (@x, @y).
 * 
 * Return value: the #NxpCanvasObject if found, or %NULL if not found.
 **/
NxpCanvasObject*
nxp_canvas_container_get_child_at_point(NxpCanvasContainer *self,
					gint x, gint y)
{
    NxpCanvasObject *rv = NULL;
    g_return_val_if_fail(self, NULL);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NULL);
    g_signal_emit(G_OBJECT(self), signals[GET_CHILD_AT_POINT_SIGNAL], 0, x, y, &rv);
    if (rv != (NxpCanvasObject *) self)
	return rv;
    else
	return NULL;
}


/**
 * nxp_canvas_container_get_children:
 * @self: the container
 * 
 * Get a list of all objects directly inside the container.
 * 
 * Return value: a #GSList with all children.
 **/
GSList*
nxp_canvas_container_get_children(NxpCanvasContainer *self)
{
    GSList *rv = NULL;
    g_return_val_if_fail(self, NULL);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NULL);
    g_signal_emit(G_OBJECT(self), signals[GET_CHILDREN_SIGNAL], 0, &rv);
    return rv;
}


static gboolean nxp_canvas_container_button_press(NxpCanvasObject *obj, GdkEvent *ev1)
{
/*     GdkEventButton *event = (GdkEventButton *) ev1; */
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(obj);

    /* If there is a pointer grab going on, we immediatly know which
       object should receive this event. */
    if (self->grab_object) {
	return nxp_canvas_object_button_press(self->grab_object, ev1);
    }
    /* Otherwise, forward the event to the object under the pointer. */
    if (self->in_object)
	return nxp_canvas_object_button_press(self->in_object, ev1);
    return FALSE;
}


static gboolean nxp_canvas_container_button_release(NxpCanvasObject *obj, GdkEvent *ev1)
{
/*     GdkEventButton *event = (GdkEventButton *) ev1; */
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(obj);
    /* If there is a pointer grab going on, we immediatly know which
       object should receive this event. */
    if (self->grab_object) {
	return nxp_canvas_object_button_release(self->grab_object, ev1);
    }
    /* Otherwise, forward the event to the object under the pointer. */
    if (self->in_object)
	return nxp_canvas_object_button_release(self->in_object, ev1);
    return FALSE;
}


static gboolean nxp_canvas_container_motion_notify(NxpCanvasObject *object, GdkEvent *ev1)
{
    GdkEventMotion *event = (GdkEventMotion *) ev1;
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(object);
    NxpCanvasObject *obj;
    int x, y;
    GdkModifierType state;
    
    x = event->x;
    y = event->y;
    state = event->state;
    /* If there is a pointer grab going on, we immediatly know which
       object should receive this event. */
    if (self->grab_object)
	return nxp_canvas_object_motion_notify(self->grab_object, ev1);

    /* Check to see if the pointer is still over the last object */
/*     if (self->in_object) { */
/* 	if (gdk_rectangle_point_in(&self->in_object->wbbox, x, y)) { */
/* 	    nxp_canvas_object_motion_notify(self->in_object, ev1); */
/* 	    return; */
/* 	} else { */
/* 	    obj = self->in_object; */
/* 	    self->in_object = NULL; */
/* 	    nxp_canvas_object_leave(obj); */
/* 	} */
/*     } */
/*     g_assert(!self->in_object); */
    /* See if the pointer is over any object. */
    obj = nxp_canvas_container_get_child_at_point(self, x, y);
    if (self->in_object && self->in_object != obj) {
	NxpCanvasObject *obj = self->in_object;
	self->in_object = NULL;
	nxp_canvas_object_leave(obj);
    }
    if (obj) {
	if (obj != self->in_object) {
	    self->in_object = obj;
	    nxp_canvas_object_enter(self->in_object);
	}
	return nxp_canvas_object_motion_notify(self->in_object, ev1);
    }
    return FALSE;
}


static gboolean nxp_canvas_container_key_press(NxpCanvasObject *object, GdkEvent *ev1)
{
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(object);
    NxpCanvasObject *obj = nxp_canvas_container_get_focus_object(self);
    if (obj) return nxp_canvas_object_key_press(obj, ev1);
    return FALSE;
}


static void nxp_canvas_container_update(NxpCanvasObject *object)
{
    NxpCanvasContainer *self = NXP_CANVAS_CONTAINER(object);
    GSList *l;

    g_return_if_fail(self);
    l = nxp_canvas_container_get_children(self);
    if (!l) return;
    for (; l; l = l->next)
	nxp_canvas_object_update(NXP_CANVAS_OBJECT(l->data));
}


void nxp_canvas_container_check_mouse_over(NxpCanvasContainer *self, gint x, gint y)
{
    NxpCanvasObject *obj;

    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(self));
    obj = nxp_canvas_container_get_child_at_point(self, x, y);
    if (obj != self->in_object) {
	if (self->in_object) nxp_canvas_object_leave(self->in_object);
	if (obj) nxp_canvas_object_enter(obj);
	self->in_object = obj;
    }
    if (self->in_object && self->in_object->interaction_mask)
	nxp_canvas_overlay_show(self->in_object->bbox_overlay);
}


void nxp_canvas_container_pointer_grab(NxpCanvasContainer *self, NxpCanvasObject *object,
				       gboolean recursive)
{
    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(self));
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(object));
    g_return_if_fail(!self->grab_object);
    g_return_if_fail(object->container == self);
    self->grab_object = object;
    if (self->in_object && self->in_object != object)
	nxp_canvas_object_leave(self->in_object);
    if (self->in_object != object) {
	self->in_object = object;
	nxp_canvas_object_enter(self->in_object);
    }
    self->grab_is_recursive = recursive;
    if (recursive && NXP_CANVAS_OBJECT(self)->container)
	nxp_canvas_container_pointer_grab(NXP_CANVAS_OBJECT(self)->container,
					  NXP_CANVAS_OBJECT(self), TRUE);
}


void nxp_canvas_container_pointer_ungrab(NxpCanvasContainer *self)
{
    gint x, y;
    GdkModifierType mask;
    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(self));
    self->grab_object = NULL;
    nxp_canvas_get_pointer(NXP_CANVAS(NXP_CANVAS_OBJECT(self)->canvas), &x, &y, &mask);
    nxp_canvas_container_check_mouse_over(self, x, y);
    if (self->grab_is_recursive && NXP_CANVAS_OBJECT(self)->container)
	nxp_canvas_container_pointer_ungrab(NXP_CANVAS_OBJECT(self)->container);

}


NxpCanvasContainer* nxp_canvas_container_new(void);

/* Don't use this! This class is really abstract, but this function is
 * needed for language bindings. */
NxpCanvasContainer* nxp_canvas_container_new(void)
{
    return g_object_new(NXP_TYPE_CANVAS_CONTAINER, NULL);
}


