/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
/* This file is based on gtkfixed.c */

#include "nxpcanvasfixed.h"


static GSList*          nxp_canvas_fixed_get_children       (NxpCanvasContainer         *self);
static NxpCanvasObject* nxp_canvas_fixed_get_child_at_point (NxpCanvasContainer         *self,
							     gint                        x,
							     gint                        y);
static void             nxp_canvas_fixed_class_init         (NxpCanvasFixedClass        *klass);
static void             nxp_canvas_fixed_init               (NxpCanvasFixed             *fixed);
static NxplotPoint*     nxp_canvas_fixed_size_request       (NxpCanvasObject            *object);
static void             nxp_canvas_fixed_configure          (NxpCanvasObject            *self,
							     NxplotRect                 *new_bbox,
							     NxpCanvasObjectInteraction  inter);
static void             nxp_canvas_fixed_shutdown           (GObject                    *obj);



static NxpCanvasContainerClass *parent_class = NULL;


static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxpCanvasFixedClass), 

    NULL, /* (GBaseInitFunc) */
    NULL, // GBaseFinalizeFunc      base_finalize;

    /* classed types, instantiated types */
    (GClassInitFunc)nxp_canvas_fixed_class_init, 
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;

    /* instantiated types */
    sizeof(NxpCanvasFixed),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxp_canvas_fixed_init, //GInstanceInitFunc      instance_init;

    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};


GType nxp_canvas_fixed_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(NXP_TYPE_CANVAS_CONTAINER, "NxpCanvasFixed",
				      &typeinfo, 0);
    }
    return type;
}


static void nxp_canvas_fixed_class_init(NxpCanvasFixedClass *class)
{
    GObjectClass            *object_class    = G_OBJECT_CLASS(class);
    NxpCanvasContainerClass *container_class = NXP_CANVAS_CONTAINER_CLASS(class);
    NxpCanvasObjectClass    *cobj_class      = NXP_CANVAS_OBJECT_CLASS(class);

    parent_class = g_type_class_peek_parent (class);

    object_class->finalize              = nxp_canvas_fixed_shutdown;
    container_class->get_child_at_point = nxp_canvas_fixed_get_child_at_point;
    container_class->get_children       = nxp_canvas_fixed_get_children;
    cobj_class->size_request            = nxp_canvas_fixed_size_request;
    cobj_class->configure               = nxp_canvas_fixed_configure;
}


static void nxp_canvas_fixed_init(NxpCanvasFixed *self)
{
    NxpCanvasObject *obj = NXP_CANVAS_OBJECT(self);
    self->children = NULL;
    obj->drawable = FALSE;
    obj->visible = TRUE;
}


NxpCanvasObject* nxp_canvas_fixed_new(void)
{
    GTypeInstance *instance;
    NxpCanvasFixed *self;

    instance = g_type_create_instance(nxp_canvas_fixed_get_type());
    self = NXP_CANVAS_FIXED(instance);

    return NXP_CANVAS_OBJECT(self);
}



void
nxp_canvas_fixed_put(NxpCanvasFixed *fixed, NxpCanvasObject *object, double x, double y)
{
    NxpCanvasFixedChild *child_info;

    g_return_if_fail(NXP_IS_CANVAS_FIXED(fixed));
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(object));

    child_info = g_new(NxpCanvasFixedChild, 1);
    child_info->object = object;
    child_info->x = x;
    child_info->y = y;

    fixed->children = g_list_append(fixed->children, child_info);
    nxp_canvas_object_set_container(object, NXP_CANVAS_CONTAINER(fixed));
    nxp_canvas_fixed_configure(NXP_CANVAS_OBJECT(fixed), 
			       &NXP_CANVAS_OBJECT(fixed)->bbox,
			       NXP_COBJ_INTERACTION_MOVE_RESIZE);
}

static void
nxp_canvas_fixed_configure(NxpCanvasObject *object,
			   NxplotRect *bbox,
			   NxpCanvasObjectInteraction inter)
{
    NxpCanvasFixed *self = NXP_CANVAS_FIXED(object);
    NxpCanvasFixedChild *child;
    NxplotPoint *req;
    double width, height;
    GList *l;
    for (l = self->children; l; l = g_list_next(l))
    {
	child = (NxpCanvasFixedChild*)(l->data);
	width = height = 0;
	req = nxp_canvas_object_size_request(child->object);
	if (!req) continue;
	child->bbox.x1 = child->x;
	child->bbox.y1 = child->y;
	child->bbox.x2 = child->x + req->x;
	child->bbox.y2 = child->y + req->y;
	g_boxed_free(TYPE_NXPLOT_POINT, req);
	nxp_canvas_object_configure(child->object, &child->bbox, 
				    NXP_COBJ_INTERACTION_MOVE_RESIZE);
	nxp_canvas_object_queue_draw(child->object);
    }
    
}

void
nxp_canvas_fixed_move(NxpCanvasFixed       *fixed, 
		      NxpCanvasObject      *object, 
		      double x, double y)
{
    /* TODO */
}



static NxplotPoint*
nxp_canvas_fixed_size_request(NxpCanvasObject *object)
{
    NxpCanvasFixed *fixed;
    NxpCanvasFixedChild *child;
    GList *children;
    gdouble xmin = G_MAXDOUBLE, xmax = G_MINDOUBLE;
    gdouble ymin = G_MAXDOUBLE, ymax = G_MINDOUBLE;
    double child_width, child_height;
    NxplotRect child_bbox;
    NxplotPoint *rv, *req;
    fixed = NXP_CANVAS_FIXED(object);
    children = fixed->children;
    while (children) {
	child = children->data;
	children = children->next;
	if (NXP_CANVAS_OBJECT_VISIBLE(child->object)) {
	    child_width = child_height = 0;
	    req = nxp_canvas_object_size_request(child->object);
	    if (!req) continue;
	    child_bbox.x1 = child->x;
	    child_bbox.y1 = child->y;
	    child_bbox.x2 = child->x + req->x;
	    child_bbox.y2 = child->y + req->y;
	    g_free(req); //g_boxed_free(TYPE_NXPLOT_POINT, req);
	    xmin = MIN(xmin, child_bbox.x1);
	    xmax = MAX(xmax, child_bbox.x2);
	    ymin = MIN(ymin, child_bbox.y1);
	    ymax = MAX(ymax, child_bbox.y2);
	}
    }
    rv = g_new(NxplotPoint, 1);
    rv->x = xmax - xmin;
    rv->y = ymax - ymin;
    return rv;
}


static void nxp_canvas_fixed_shutdown(GObject *obj)
{
    NxpCanvasFixed *self = NXP_CANVAS_FIXED(obj);
    NxpCanvasFixedChild *child;

    while (self->children) {
	child = (NxpCanvasFixedChild *) self->children->data;
	g_object_unref(child->object);
	g_free(child);
	self->children = g_list_delete_link(self->children, self->children);
    }
}


static GSList* nxp_canvas_fixed_get_children(NxpCanvasContainer *container)
{
    GSList *new = NULL;
    GList *our;
    NxpCanvasFixed *self = NXP_CANVAS_FIXED(container);
    for (our = self->children; our; our = g_list_next(our)) 
	new = g_slist_append(new, ((NxpCanvasFixedChild *) our->data)->object);
    return new;
}


static NxpCanvasObject*
nxp_canvas_fixed_get_child_at_point(NxpCanvasContainer *container, gint x, gint y)
{
    NxpCanvasFixed *self = NXP_CANVAS_FIXED(container);
    NxpCanvasObject *child;
    GList *list;

    for (list = self->children; list; list = g_list_next(list)) {
	child = ((NxpCanvasFixedChild *) list->data)->object;
	if (gdk_rectangle_point_in(&child->wbbox, x, y))
	    return child;
    }
    return NULL;
}

