/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#ifndef __NXP_CANVAS_OBJECT_H__
#define __NXP_CANVAS_OBJECT_H__

#include <glib-object.h>

#define NXP_CANVAS_OBJECT(object)            (G_TYPE_CHECK_INSTANCE_CAST ((object),\
				nxp_canvas_object_get_type(), NxpCanvasObject))

#define NXP_CANVAS_OBJECT_CLASS(klass)	     (G_TYPE_CHECK_CLASS_CAST ((klass),\
				nxp_canvas_object_get_type(), NxpCanvasObjectClass))

#define NXP_IS_CANVAS_OBJECT(object)         (G_TYPE_CHECK_INSTANCE_TYPE ((object),\
				nxp_canvas_object_get_type()))

#define NXP_IS_CANVAS_OBJECT_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE ((klass),\
				nxp_canvas_object_get_type()))

#define NXP_CANVAS_OBJECT_GET_CLASS(object)  (G_TYPE_INSTANCE_GET_CLASS ((object),\
				nxp_canvas_object_get_type(), NxpCanvasObjectClass))

#define NXP_CANVAS_OBJECT_TYPE(object)       (G_TYPE_FROM_INSTANCE (object))
#define NXP_CANVAS_OBJECT_TYPE_NAME(object)  (g_type_name (NXP_CANVAS_OBJECT_TYPE (object)))
#define NXP_CANVAS_OBJECT_CLASS_TYPE(klass)  (G_TYPE_FROM_CLASS (klass))
#define NXP_CANVAS_OBJECT_CLASS_NAME(klass)  (g_type_name (NXP_CANVAS_OBJECT_CLASS_TYPE (klass)))

#define NXP_TYPE_CANVAS_OBJECT    (nxp_canvas_object_get_type())

#define NXP_CANVAS_OBJECT_VISIBLE(obj) (NXP_CANVAS_OBJECT(obj)->drawable &&\
                                        NXP_CANVAS_OBJECT(obj)->visible)


typedef struct _NxpCanvasObject      NxpCanvasObject;
typedef struct _NxpCanvasObjectClass NxpCanvasObjectClass;
/* typedef struct _NxpCanvasContainer   NxpCanvasContainer; */


#include "nxpcanvas.h"
#include "nxpcanvasoverlay.h"
/* #include "nxpcanvascontainer.h" */


typedef enum {
    NXP_COBJ_INTERACTION_NONE,
    NXP_COBJ_INTERACTION_MOVE,
    NXP_COBJ_INTERACTION_RESIZE_N,
    NXP_COBJ_INTERACTION_RESIZE_NE,
    NXP_COBJ_INTERACTION_RESIZE_E,
    NXP_COBJ_INTERACTION_RESIZE_SE,
    NXP_COBJ_INTERACTION_RESIZE_S,
    NXP_COBJ_INTERACTION_RESIZE_SW,
    NXP_COBJ_INTERACTION_RESIZE_W,
    NXP_COBJ_INTERACTION_RESIZE_NW,
    NXP_COBJ_INTERACTION_MOVE_RESIZE,
} NxpCanvasObjectInteraction;

#define nxp_canvas_object_interaction_is_resize(i)\
	((i) == NXP_COBJ_INTERACTION_RESIZE_N  ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_NE ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_E  ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_SE ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_S  ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_SW ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_W  ||\
	 (i) == NXP_COBJ_INTERACTION_RESIZE_NW ||\
         (i) == NXP_COBJ_INTERACTION_MOVE_RESIZE)


typedef enum {
    NXP_COBJ_INTERACTION_MOVE_MASK 	= (1<<0),
    NXP_COBJ_INTERACTION_RESIZE_N_MASK 	= (1<<1),
    NXP_COBJ_INTERACTION_RESIZE_NE_MASK = (1<<2),
    NXP_COBJ_INTERACTION_RESIZE_E_MASK 	= (1<<3),
    NXP_COBJ_INTERACTION_RESIZE_SE_MASK = (1<<4),
    NXP_COBJ_INTERACTION_RESIZE_S_MASK 	= (1<<5),
    NXP_COBJ_INTERACTION_RESIZE_SW_MASK = (1<<6),
    NXP_COBJ_INTERACTION_RESIZE_W_MASK 	= (1<<7),
    NXP_COBJ_INTERACTION_RESIZE_NW_MASK = (1<<8)
} NxpCanvasObjectInteractionMask;



struct _NxpCanvasObject
{
    GObject base;
    NxpCanvas *canvas;		/* parent NxpCanvas */
    NxpCanvasContainer *container; /* parent container */
    NxplotRect bbox;		/* bounding box in user coordinates */
    GdkRectangle wbbox;	        /* bounding box in window coordinates */
    NxpCanvasOverlay *bbox_overlay; /* A reference to the object's bbox overlay */
    NxpCanvasOverlay *ghost_bbox; /* A bbox indicating the old
				     position of the object when moving */
    GdkPoint drag_origin;	/* Point where move/resize drag has begun */
    NxpCanvasObjectInteraction interaction; /* current interaction underway */
    NxpCanvasObjectInteractionMask interaction_mask; /* allowed interactions */
    gboolean whole_area_interaction:1;
    gboolean dirty:1;		/* flag that indicates if it needs
				   redrawing (to be managed by the
				   canvas) */
    gboolean drawable:1;	/* Some objects, such as some containers, don't draw anything. */
    gboolean visible:1;		/* Even drawable objects may be temporarily invisible */
    gulong canvas_notify_sig;
};


struct _NxpCanvasObjectClass
{
    GObjectClass base;

    void         (*draw)	   (NxpCanvasObject *self, Nxplot *pl);
    void         (*update)	   (NxpCanvasObject *self);
    void         (*enter)	   (NxpCanvasObject *self);
    void         (*leave)	   (NxpCanvasObject *self);
    gboolean     (*button_press)   (NxpCanvasObject *self, GdkEvent *event);
    gboolean     (*button_release) (NxpCanvasObject *self, GdkEvent *event);
    gboolean     (*motion_notify)  (NxpCanvasObject *self, GdkEvent *event);
    gboolean     (*key_press)	   (NxpCanvasObject *self, GdkEvent *event);
    NxplotPoint* (*size_request)   (NxpCanvasObject *self);
    void         (*configure)      (NxpCanvasObject *self, NxplotRect *new_bbox,
			 	    NxpCanvasObjectInteraction inter);
};


#include "nxpcanvascontainer.h"


GType               nxp_canvas_object_get_type       (void);
void                nxp_canvas_object__set_visible   (NxpCanvasObject            *obj,
						      gboolean                    visible);
void                nxp_canvas_object__set_drawable  (NxpCanvasObject            *obj,
						      gboolean                    drawable);
void                nxp_canvas_object_draw           (NxpCanvasObject            *self,
						      Nxplot                     *pl);
void                nxp_canvas_object_update_wbbox   (NxpCanvasObject            *self);
void                nxp_canvas_object_set_canvas     (NxpCanvasObject            *self,
						      NxpCanvas                  *canvas);
void                nxp_canvas_object_set_container  (NxpCanvasObject            *self,
						      NxpCanvasContainer         *container);
void                nxp_canvas_object_enter          (NxpCanvasObject            *self);
void                nxp_canvas_object_leave          (NxpCanvasObject            *self);
void                nxp_canvas_object_update         (NxpCanvasObject            *self);
void                nxp_canvas_object_queue_draw     (NxpCanvasObject            *self);
gboolean            nxp_canvas_object_button_press   (NxpCanvasObject            *self,
						      GdkEvent                   *event);
gboolean            nxp_canvas_object_button_release (NxpCanvasObject            *self,
						      GdkEvent                   *event);
gboolean            nxp_canvas_object_motion_notify  (NxpCanvasObject            *self,
						      GdkEvent                   *event);
gboolean            nxp_canvas_object_key_press      (NxpCanvasObject            *self,
						      GdkEvent                   *event);
void                nxp_canvas_object_configure      (NxpCanvasObject            *self,
						      NxplotRect                 *new_bbox,
						      NxpCanvasObjectInteraction  inter);
NxplotPoint*        nxp_canvas_object_size_request   (NxpCanvasObject            *self);
Nxplot*             nxp_canvas_object_get_plotter    (NxpCanvasObject            *self);
NxpCanvasContainer* nxp_canvas_object_get_toplevel   (NxpCanvasObject            *self);



#endif /* __NXP_CANVAS_OBJECT_H__ */
