/* -*- Mode: C; c-file-style: "bsd"; -*- */
#include "nxpcanvastextcursor.h"
#include <string.h>

/*  static guint signals[LAST_SIGNAL] = { 0 }; */
/*  static NxpCanvasTextCursorClass *nxp_canvas_text_cursor_class = NULL; */
static gpointer parent_class = NULL;

enum {
    PROP_0,
    PROP_LINE_STYLE,
    PROP_WIDTH,
    PROP_DELAY,
    PROP_P1X,
    PROP_P1Y,
    PROP_P2X,
    PROP_P2Y,
    PROP_GHOST
};


static void     nxp_canvas_text_cursor_init            (NxpCanvasTextCursor      *instance,
							gpointer                  g_class);
static void     nxp_canvas_text_cursor_class_base_init (NxpCanvasTextCursorClass *class);
static void     nxp_canvas_text_cursor_class_init      (NxpCanvasTextCursorClass *class,
							gpointer                  data);
static void     nxp_canvas_text_cursor_shutdown        (GObject                  *object);
static void     nxp_canvas_text_cursor_draw            (NxpCanvasOverlay         *self,
							GdkDrawable              *drawable,
							GdkGC                    *gc,
							guint                     t);
static void     get_property                           (GObject                  *object,
							guint                     prop_id,
							GValue                   *value,
							GParamSpec               *pspec);
static void     set_property                           (GObject                  *object,
							guint                     prop_id,
							const GValue             *value,
							GParamSpec               *pspec);
static gboolean timeout_function                       (NxpCanvasTextCursor      *self);




static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxpCanvasTextCursorClass), 

    (GBaseInitFunc)nxp_canvas_text_cursor_class_base_init, 
    NULL, // GBaseFinalizeFunc      base_finalize;

    /* classed types, instantiated types */
    (GClassInitFunc)nxp_canvas_text_cursor_class_init, 
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;

    /* instantiated types */
    sizeof(NxpCanvasTextCursor),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxp_canvas_text_cursor_init, //GInstanceInitFunc      instance_init;

    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};


GType nxp_canvas_text_cursor_get_type(void)
{
    static GType type = 0;
    if (!type) 
    {
	type = g_type_register_static(NXP_TYPE_CANVAS_OVERLAY, "NxpCanvasTextCursor",
				      &typeinfo, 0);
    }
    return type;
}


static void nxp_canvas_text_cursor_init(NxpCanvasTextCursor *self, gpointer g_class)
{
/*      if (!nxp_canvas_text_cursor_class) */
/*  	nxp_canvas_text_cursor_class = NXP_CANVAS_TEXT_CURSOR_GET_CLASS(self); */
    NxpCanvasOverlay *overlay = NXP_CANVAS_OVERLAY(self);
    overlay->bbox.x = 0;
    overlay->bbox.y = 0;
    overlay->bbox.width = 0;
    overlay->bbox.height = 0;

}


static void nxp_canvas_text_cursor_class_base_init(NxpCanvasTextCursorClass *class)
{
    /* empty */
}


static void nxp_canvas_text_cursor_class_init(NxpCanvasTextCursorClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS(class);
    NxpCanvasOverlayClass *overlay_class = NXP_CANVAS_OVERLAY_CLASS(class);

    parent_class = g_type_class_peek_parent (class);
    object_class->set_property = set_property;
    object_class->get_property = get_property;
    object_class->finalize = nxp_canvas_text_cursor_shutdown;
    overlay_class->draw = nxp_canvas_text_cursor_draw;

    g_object_class_install_property(object_class,
				    PROP_LINE_STYLE,
				    g_param_spec_int(
					"line-style", "GdkLineStyle",
					"Style of line to draw the rectangle",
					GDK_LINE_SOLID,	/* minimum */
					GDK_LINE_DOUBLE_DASH, /* maximum */
					GDK_LINE_SOLID,	/* default */
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_WIDTH,
				    g_param_spec_int(
					"width",
					"Line width",
					"Line width in pixels",
					0, G_MAXINT, 2,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_DELAY,
				    g_param_spec_uint(
					"delay", "Blink delay",
					"Blink speed (delay)",
					0, G_MAXUINT, 2,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_P1X,
				    g_param_spec_int(
					"p1_x", "Point 1 X",
					"Point 1 X",
					G_MININT, G_MAXINT, 0,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_P1Y,
				    g_param_spec_int(
					"p1_y", "Point 1 Y",
					"Point 1 Y",
					G_MININT, G_MAXINT, 0,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_P2X,
				    g_param_spec_int(
					"p2_x", "Point 2 X",
					"Point 2 X",
					G_MININT, G_MAXINT, 0,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_P2Y,
				    g_param_spec_int(
					"p2_y", "Point 2 Y",
					"Point 2 Y",
					G_MININT, G_MAXINT, 0,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_GHOST,
				    g_param_spec_boolean(
					"ghost",
					"ghost mode on/off",
					"ghost mode on/off",
					FALSE,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
}


static void nxp_canvas_text_cursor_shutdown(GObject *object)
{
    NxpCanvasTextCursor *self;
    self = NXP_CANVAS_TEXT_CURSOR(object);
    nxp_canvas_text_cursor_stop_blink(self);
    if (G_OBJECT_CLASS(parent_class)->finalize)
	G_OBJECT_CLASS(parent_class)->finalize(object);
}


static void update_bbox(NxpCanvasTextCursor *self)
{
    NxpCanvasOverlay *overlay = NXP_CANVAS_OVERLAY(self);
    /* A bit pessimist bbox, but easier to calculate */
    overlay->bbox.x = MIN(self->p1.x, self->p2.x) - self->width;
    overlay->bbox.y = MIN(self->p1.y, self->p2.y) - self->width;
    overlay->bbox.width = MAX(self->p1.x, self->p2.x) - overlay->bbox.x + self->width;
    overlay->bbox.height = MAX(self->p1.y, self->p2.y) - overlay->bbox.x + self->width;
}

static void nxp_canvas_text_cursor_draw(NxpCanvasOverlay *overlay, GdkDrawable *drawable,
					GdkGC *gc, guint t)
{
    NxpCanvasTextCursor *self = NXP_CANVAS_TEXT_CURSOR(overlay);
    gint8 dashes[] = {self->width};
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_TEXT_CURSOR(self));

    gdk_gc_set_dashes(gc, 0, dashes, 1);
    gdk_gc_set_line_attributes(gc, self->width, self->ghost? GDK_LINE_ON_OFF_DASH : 
			       self->line_style, GDK_CAP_NOT_LAST,
			       GDK_JOIN_MITER);
    gdk_draw_line(drawable, gc, self->p1.x, self->p1.y, self->p2.x, self->p2.y);
}





static void
set_property (GObject      *object,
	      guint         prop_id,
	      const GValue *value,
	      GParamSpec   *pspec)
{
    NxpCanvasTextCursor *self;
    NxpCanvasOverlay *overlay;

    self = NXP_CANVAS_TEXT_CURSOR(object);
    overlay = NXP_CANVAS_OVERLAY(object);

    switch (prop_id)
    {

    case PROP_LINE_STYLE: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->line_style = g_value_get_int(value);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_GHOST: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->ghost = g_value_get_boolean(value);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_WIDTH: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->width = g_value_get_int(value);
	update_bbox(self);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_DELAY: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->delay = g_value_get_uint(value);
	if (self->timeout_id) {
	    gtk_timeout_remove(self->timeout_id);
	    self->timeout_id = 
		gtk_timeout_add(self->delay, (GtkFunction)timeout_function, self);
	}
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }
	
    case PROP_P1X: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->p1.x = g_value_get_int(value);
	update_bbox(self);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_P1Y: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->p1.y = g_value_get_int(value);
	update_bbox(self);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_P2X: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->p2.x = g_value_get_int(value);
	update_bbox(self);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    case PROP_P2Y: {
	gboolean visible;
	if (overlay->canvas && overlay->visible) {
	    nxp_canvas_overlay_hide(overlay);
	    visible = TRUE;
	} else visible = FALSE;
	self->p2.y = g_value_get_int(value);
	update_bbox(self);
	if (visible)
	    nxp_canvas_overlay_show(overlay);
	break;
    }

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}



static void
get_property (GObject      *object,
                           guint         prop_id,
                           GValue       *value,
                           GParamSpec   *pspec)
{
    NxpCanvasTextCursor *self;

    self = NXP_CANVAS_TEXT_CURSOR(object);

    switch (prop_id)
    {
    case PROP_LINE_STYLE:
	g_value_set_int(value, self->line_style);
	break;
    case PROP_WIDTH:
	g_value_set_int(value, self->width);
	break;
    case PROP_DELAY:
	g_value_set_uint(value, self->delay);
	break;
    case PROP_P1X:
	g_value_set_int(value, self->p1.x);
	break;
    case PROP_P1Y:
	g_value_set_int(value, self->p1.y);
	break;
    case PROP_P2X:
	g_value_set_int(value, self->p2.x);
	break;
    case PROP_P2Y:
	g_value_set_int(value, self->p2.y);
	break;
    case PROP_GHOST:
	g_value_set_boolean(value, self->ghost);
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}


static gboolean timeout_function(NxpCanvasTextCursor *self)
{
    NxpCanvasOverlay *overlay = NXP_CANVAS_OVERLAY(self);
    /* Blink! :) */
    if (overlay->visible)
	nxp_canvas_overlay_hide(overlay);
    else
	nxp_canvas_overlay_show(overlay);
    return TRUE;
}


void nxp_canvas_text_cursor_start_blink(NxpCanvasTextCursor *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_TEXT_CURSOR(self));
    if (self->timeout_id) return;
    self->timeout_id = 
	gtk_timeout_add(self->delay, (GtkFunction)timeout_function, self);
}


void nxp_canvas_text_cursor_stop_blink(NxpCanvasTextCursor *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_TEXT_CURSOR(self));
    if (!self->timeout_id) return;
    gtk_timeout_remove(self->timeout_id);
    self->timeout_id = 0;
}


