/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#define _GNU_SOURCE
#include "nxplot-gp.h"
#include <gdk/gdk.h>
#include <string.h>
#include <libart_lgpl/art_affine.h>
#include <math.h>
#include "config.h"

static NxplotGPClass *nxplot_gp_class = NULL;

static GnomeGlyphList* nxplot_gp_text_layout(NxplotGP *self, const char *text);


static void   nxplot_gp_init            (NxplotGP           *instance,
					 gpointer            g_class);
static void   nxplot_gp_class_base_init (NxplotGPClass      *class);
static void   nxplot_gp_class_init      (NxplotGPClass      *class,
					 gpointer            data);
static void   nxplot_gp_shutdown        (GObject            *object);
static void   nxplot_gp_open            (Nxplot             *nxplot);
static void   nxplot_gp_close           (Nxplot             *nxplot);
static void   nxplot_gp_erase           (Nxplot             *nxplot);
static void   nxplot_gp_line            (Nxplot             *nxplot,
					 double              x1,
					 double              y1,
					 double              x2,
					 double              y2);
static void   nxplot_gp_linerel         (Nxplot             *nxplot,
					 double              x1,
					 double              y1,
					 double              x2,
					 double              y2);
static void   nxplot_gp_linewidth       (Nxplot             *nxplot,
					 double              width);
static void   nxplot_gp_pencolor        (Nxplot             *nxplot,
					 guint16             red,
					 guint16             green,
					 guint16             blue);
static void   nxplot_gp_bgcolor         (Nxplot             *nxplot,
					 guint16             red,
					 guint16             green,
					 guint16             blue);
static void   nxplot_gp_pencolorname    (Nxplot             *nxplot,
					 const char         *name);
static void   nxplot_gp_move            (Nxplot             *nxplot,
					 double              x,
					 double              y);
static void   nxplot_gp_moverel         (Nxplot             *nxplot,
					 double              x,
					 double              y);
static void   nxplot_gp_fontsize        (Nxplot             *nxplot,
					 double              size);
static void   nxplot_gp_textangle       (Nxplot             *nxplot,
					 double              angle);
static void   nxplot_gp_linemode        (Nxplot             *nxplot,
					 NxplotLinemode      mode);
static void   nxplot_gp_label           (Nxplot             *nxplot,
					 NxplotHAlign        halign,
					 NxplotVAlign        valign,
					 const char         *text);
static void   nxplot_gp_labelmetrics    (Nxplot             *nxplot,
					 const char         *text,
					 NxplotLabelMetrics *metrics);
static double nxplot_gp_labelwidth      (Nxplot             *nxplot,
					 const char         *text);
static void   nxplot_gp_endpath         (Nxplot             *nxplot);
static void   nxplot_gp_space           (Nxplot             *nxplot,
					 double              x1,
					 double              y1,
					 double              x2,
					 double              y2);
static void   nxplot_gp_circle          (Nxplot             *nxplot,
					 double              xc,
					 double              yc,
					 double              r);
static void   nxplot_gp_savestate       (Nxplot             *nxplot);
static void   nxplot_gp_restorestate    (Nxplot             *nxplot);
static void   nxplot_gp_setmatrix       (Nxplot             *nxplot,
					 double              m0,
					 double              m1,
					 double              m2,
					 double              m3,
					 double              tx,
					 double              ty);
static void   nxplot_gp_concat          (Nxplot             *nxplot,
					 double              m0,
					 double              m1,
					 double              m2,
					 double              m3,
					 double              tx,
					 double              ty);
static void   nxplot_gp_rotate          (Nxplot             *nxplot,
					 double              theta);
static void   nxplot_gp_scale           (Nxplot             *nxplot,
					 double              sx,
					 double              sy);
static void   nxplot_gp_translate       (Nxplot             *nxplot,
					 double              tx,
					 double              ty);
static void   nxplot_gp_fontname        (Nxplot             *nxplot,
					 const char         *font_name);
static double nxplot_gp_descender       (Nxplot             *nxplot);
static void   nxplot_gp_setcliprect     (Nxplot             *nxplot,
					 NxplotRect         *clip_rect);
static void   nxplot_gp_clear_states    (Nxplot             *nxplot);
static void   nxplot_gp_fillcolor       (Nxplot             *nxplot,
					 guint16             red,
					 guint16             green,
					 guint16             blue);
static void   nxplot_gp_fillcolorname   (Nxplot             *nxplot,
					 const char         *name);
static void   nxplot_gp_fillopacity     (Nxplot             *nxplot,
					 guint16             opacity);




static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxplotGPClass),

    (GBaseInitFunc)nxplot_gp_class_base_init,
    NULL, // GBaseFinalizeFunc      base_finalize;
    
    /* classed types, instantiated types */
    (GClassInitFunc)nxplot_gp_class_init,
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;
    
    /* instantiated types */
    sizeof(NxplotGP),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxplot_gp_init, //GInstanceInitFunc      instance_init;
    
    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};



GType nxplot_gp_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(nxplot_get_type(), "NxplotGP", 
				      &typeinfo, 0);
    }
    return type;
}


static void nxplot_gp_init(NxplotGP *self, gpointer g_class)
{
    if (!nxplot_gp_class)
	nxplot_gp_class = NXPLOT_GP_GET_CLASS(self);
    
    self->saved_states = g_queue_new();
    self->state.linewidth = 1;
    self->font = NULL;
    self->state.font_name = g_quark_from_static_string("Sans Regular");
    art_affine_identity(self->state.ctm);
}


static void nxplot_gp_class_base_init(NxplotGPClass *class)
{
    /* empty */
}


static void nxplot_gp_class_init(NxplotGPClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    NxplotClass *nxplot_class = NXPLOT_CLASS(class);

    object_class->finalize	= nxplot_gp_shutdown;
    nxplot_class->open		= nxplot_gp_open;
    nxplot_class->close		= nxplot_gp_close;
    nxplot_class->erase		= nxplot_gp_erase;
    nxplot_class->line		= nxplot_gp_line;
    nxplot_class->linerel	= nxplot_gp_linerel;
    nxplot_class->linewidth	= nxplot_gp_linewidth;
    nxplot_class->pencolor	= nxplot_gp_pencolor;
    nxplot_class->bgcolor	= nxplot_gp_bgcolor;
    nxplot_class->pencolorname	= nxplot_gp_pencolorname;
    nxplot_class->move		= nxplot_gp_move;
    nxplot_class->moverel	= nxplot_gp_moverel;
    nxplot_class->linemode	= nxplot_gp_linemode;
    nxplot_class->textangle	= nxplot_gp_textangle;
    nxplot_class->fontname	= nxplot_gp_fontname;
    nxplot_class->fontsize	= nxplot_gp_fontsize;
    nxplot_class->label	        = nxplot_gp_label;
    nxplot_class->labelmetrics  = nxplot_gp_labelmetrics;
    nxplot_class->labelwidth    = nxplot_gp_labelwidth;
    nxplot_class->space		= nxplot_gp_space;
    nxplot_class->endpath	= nxplot_gp_endpath;
    nxplot_class->circle	= nxplot_gp_circle;
    nxplot_class->savestate	= nxplot_gp_savestate;
    nxplot_class->restorestate	= nxplot_gp_restorestate;
    nxplot_class->setmatrix	= nxplot_gp_setmatrix;
    nxplot_class->concat	= nxplot_gp_concat;
    nxplot_class->rotate	= nxplot_gp_rotate;
    nxplot_class->scale		= nxplot_gp_scale;
    nxplot_class->translate	= nxplot_gp_translate;
    nxplot_class->descender	= nxplot_gp_descender;
    nxplot_class->setcliprect	= nxplot_gp_setcliprect;
    nxplot_class->fillcolor	= nxplot_gp_fillcolor;
    nxplot_class->fillcolorname	= nxplot_gp_fillcolorname;
    nxplot_class->fillopacity	= nxplot_gp_fillopacity;
}



static void nxplot_gp_shutdown(GObject *object)
{
    Nxplot   *nxplot = NXPLOT(object);
    NxplotGP *self   = NXPLOT_GP(object);

    /* g_message(__FUNCTION__); */

    if (nxplot->opened)
	nxplot_gp_close(nxplot);
    if (self->pc) {
	g_object_unref(self->pc);
	self->pc = NULL;
    }
    if (self->font) {
	g_object_unref(self->font);
	self->font = NULL;
    }
}



Nxplot* nxplot_gp_new(GnomePrintContext *pc)
{
    GTypeInstance *instance;
    NxplotGP *self;
    Nxplot *nxplot;

    g_return_val_if_fail(GNOME_IS_PRINT_CONTEXT(pc), NULL);
    instance = g_type_create_instance(nxplot_gp_get_type());
    self = NXPLOT_GP(instance);
    nxplot = NXPLOT(instance);

    self->pc = pc;
    g_object_ref(pc);

    return NXPLOT(self);
}



static void nxplot_gp_open(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    if (nxplot->opened) return;
    /* g_message(__FUNCTION__); */
    gnome_print_beginpage(self->pc, NULL);
    /* g_message("gnome_print_beginpage()"); */
    nxplot->opened = TRUE;
}


static void nxplot_gp_close(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    if (!nxplot->opened) return;
    /* g_message(__FUNCTION__); */
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    gnome_print_showpage(self->pc);
    /* g_message("gnome_print_showpage()"); */
    nxplot_gp_clear_states(nxplot);
    nxplot->opened = FALSE;
    gnome_print_context_close(self->pc);
    /* g_message("gnome_print_context_close()"); */
}


static void nxplot_gp_erase(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    gnome_print_showpage(self->pc);
    gnome_print_beginpage(self->pc, NULL);
}


static void nxplot_gp_line(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    int rv;

    g_return_if_fail(nxplot->opened);
    /* g_message(__FUNCTION__); */
    if (self->inpath) {
	if (self->state.last_x == x1 && self->state.last_y == y1) {
	    /* line does not interrupt path */
	    if ((rv = gnome_print_lineto(self->pc, x2, y2)) < 0)
		g_warning("gnome_print_lineto: %i", rv);
 	    /* else */
/* 		 g_message("gnome_print_lineto(%G, %G)", x2, y2);  */
	} else {
	    nxplot_gp_endpath(NXPLOT(self));
	    gnome_print_newpath(self->pc);
	    if ((rv = gnome_print_moveto(self->pc, x1, y1)) < 0)
		g_warning("gnome_print_moveto: %i", rv);
	    else
		/* g_message("gnome_print_moveto(%G, %G)", x1, y1); */
	    self->state.path_start_x = x1;
	    self->state.path_start_y = y1;

	    if ((rv = gnome_print_lineto(self->pc, x2, y2)) < 0)
		g_warning("gnome_print_lineto: %i", rv);
	    /* else */
		/* g_message("gnome_print_lineto(%G, %G)", x2, y2); */
	}
    } else {
	gnome_print_newpath(self->pc);
	self->state.path_start_x = x1;
	self->state.path_start_y = y1;

	if ((rv = gnome_print_moveto(self->pc, x1, y1)) < 0)
	    g_warning("gnome_print_moveto: %i", rv);
	else 
	    /* g_message("gnome_print_moveto(%G, %G)", x1, y1); */

	if ((rv = gnome_print_lineto(self->pc, x2, y2)) < 0)
	    g_warning("gnome_print_lineto: %i", rv);
	/* else  */
	    /* g_message("gnome_print_lineto(%G, %G)", x2, y2); */
    }
    self->inpath = TRUE;
    self->state.last_x = x2;
    self->state.last_y = y2;
}


static void nxplot_gp_linerel(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    nxplot_gp_line(nxplot, self->state.last_x + x1, self->state.last_y + y1,
		   self->state.last_x + x2, self->state.last_y + y2);
}


static void nxplot_gp_linewidth(Nxplot *nxplot, double width)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_gp_endpath(nxplot);
    gnome_print_setlinewidth(self->pc, width);
    self->state.linewidth = width;
}


static void nxplot_gp_pencolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    double r = red/65535.0, g = green/65535.0, b = blue/65535.0;
    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_gp_endpath(nxplot);
    /* g_message("gnome_print_setrgbcolor(%f, %f, %f);", r, g, b); */
    gnome_print_setrgbcolor(self->pc, r, g, b);
    self->state.foreground = ((red<<16)&0xff000000) |
	((green<<8)&0x00ff0000) | ((blue)&0x0000ff00);
}


static void nxplot_gp_bgcolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
/*     NxplotGP *self = NXPLOT_GP(nxplot); */

/*     g_return_if_fail(nxplot->opened); */
    /* TODO */
}


static void nxplot_gp_pencolorname(Nxplot *nxplot, const char *name)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    GdkColor color = {0,0,0,0};
    gint rv;

    /* g_message(__FUNCTION__); */
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    rv = gdk_color_parse(name, &color);
    if (!rv) g_warning("Parsing of color '%s' failed", name);
    nxplot_gp_pencolor(nxplot, color.red, color.green, color.blue);
}


static void nxplot_gp_move(Nxplot *nxplot, double x, double y)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_gp_endpath(nxplot);
    self->state.last_x = x;
    self->state.last_y = y;
    gnome_print_moveto(self->pc, x, y);
}


static void nxplot_gp_moverel(Nxplot *nxplot, double x, double y)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    nxplot_gp_move(nxplot, self->state.last_x + x, self->state.last_y + y);
}


static void update_font(NxplotGP *self)
{
    if (self->font)
	g_object_unref(self->font);
    self->font = gnome_font_find_closest(g_quark_to_string(self->state.font_name),
					 self->state.font_size);
    /* g_message("gnome_font_find_closest(\"%s\", %G): %p", */
/* 	      g_quark_to_string(self->state.font_name), */
/* 	      self->state.font_size, self->font); */
    if (self->font)
	gnome_print_setfont(self->pc, self->font);
}


static void nxplot_gp_fontname(Nxplot *nxplot, const char *font_name)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    self->state.font_name = g_quark_from_string(font_name);
    update_font(self);
}


static void nxplot_gp_fontsize(Nxplot *nxplot, double size)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    self->state.font_size = size;
    update_font(self);
}


static void nxplot_gp_textangle(Nxplot *nxplot, double angle)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    self->state.text_angle = angle;
}


static void nxplot_gp_endpath(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    /* g_message(__FUNCTION__); */
    if (self->inpath) {
	if (self->state.last_x == self->state.path_start_x &&
	    self->state.last_y == self->state.path_start_y)
	{
	    gnome_print_closepath(self->pc);
	    /* g_message("gnome_print_closepath()"); */
	}
	/* Handle fill */
	if (self->state.fillopacity > 0) {
	    gnome_print_gsave(self->pc);
	    gnome_print_setrgbcolor(self->pc,
				    self->state.fillcolor.red / 65535.0,
				    self->state.fillcolor.green / 65535.0,
				    self->state.fillcolor.blue / 65535.0);
	    gnome_print_setopacity(self->pc, self->state.fillopacity);
	    gnome_print_fill(self->pc);
	    gnome_print_grestore(self->pc);
	}
	gnome_print_stroke(self->pc);
	/* g_message("gnome_print_stroke()"); */
	self->inpath = FALSE;
    }
}


static void nxplot_gp_space(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    g_warning("nxplot_space() is not implemented (and probably never will be) in the GNOME Print backend");
}


static void nxplot_gp_circle(Nxplot *nxplot, double xc, double yc, double r)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    /* g_message(__FUNCTION__); */
    gnome_print_newpath(self->pc);
    gnome_print_moveto(self->pc, xc + r, yc);
    gnome_print_arcto(self->pc, xc, yc, r, 0, 180, 0);
    gnome_print_arcto(self->pc, xc, yc, r, 180, 360, 0);
    gnome_print_closepath(self->pc);
    gnome_print_stroke(self->pc);
}

#define MAX_NDASHES 8
static struct {
    int ndashes;
    double dashes[MAX_NDASHES];
} line_modes[7] = 
{
    {0, {0,0,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_SOLID */
    {2, {1,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_DOTTED */
    {4, {1,2,4,2,0,0,0,0}},	/* NXPLOT_LINEMODE_DOTDASHED */
    {2, {3,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_SHORTDASHED */
    {2, {6,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_LONGDASHED */
    {6, {1,2,1,2,4,2,0,0}},	/* NXPLOT_LINEMODE_DOTDOTDASHED */
    {8, {1,2,1,2,1,2,4,2}}	/* NXPLOT_LINEMODE_DOTDOTDOTDASHED */
};


static void nxplot_gp_linemode(Nxplot *nxplot, NxplotLinemode linemode)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_gp_endpath(nxplot);
    if (linemode > 0 && linemode < 7) {
	gnome_print_setdash(self->pc, line_modes[linemode].ndashes,
			    line_modes[linemode].dashes, 0);
    } else {
	/* g_message("gnome_print_setdash(self->pc, 0, NULL, 0);"); */
	gnome_print_setdash(self->pc, 0, NULL, 0);
    }
}



static void nxplot_gp_label(Nxplot *nxplot, NxplotHAlign halign,
			    NxplotVAlign valign, const char *text)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    double last_x = self->state.last_x;
    double last_y = self->state.last_y;
    double off_x = 0, off_y = 0;
    double c = 0, s = 0;
    double width;
    GnomeGlyphList *gl;
    ArtDRect bbox;

    g_return_if_fail(nxplot->opened);

    gl = nxplot_gp_text_layout(self, text);
    gnome_glyphlist_bbox(gl, NULL, 0, &bbox);

    off_x = -bbox.x0;
    width = bbox.x1 - bbox.x0;

    switch (halign)
    {
    case NXPLOT_HALIGN_LEFT:
	break;
    case NXPLOT_HALIGN_CENTER:
	off_x += -width/2;
	break;
    case NXPLOT_HALIGN_RIGHT:
	off_x += -width;
	break;
    default:
	g_warning("Invalid horizontal alignment %i", halign);
    }
    switch (valign)
    {
    case NXPLOT_VALIGN_BOTTOM:
	off_y = -gnome_font_get_descender(self->font);
	break;
    case NXPLOT_VALIGN_BASELINE:
	break;
    case NXPLOT_VALIGN_CENTER:
	off_y = -(gnome_font_get_ascender(self->font) +
		  gnome_font_get_descender(self->font))/2;
	break;
    case NXPLOT_VALIGN_CAPLINE:
	off_y = -(gnome_font_get_ascender(self->font) +
		  gnome_font_get_descender(self->font)); /* FIXME */
	break;
    case NXPLOT_VALIGN_TOP:
	off_y = -(gnome_font_get_ascender(self->font) +
		  gnome_font_get_descender(self->font));
	break;
    default:
	g_warning("Invalid vertical alignment %i", valign);
    }

    if (self->state.text_angle != 0) {
	nxplot_savestate(nxplot);
	/* (remember that transformations are in reverse order, always) */
	nxplot_translate(nxplot, last_x, last_y);      /* 3 */
	nxplot_rotate(nxplot, self->state.text_angle); /* 2 */
	nxplot_translate(nxplot, -last_x, -last_y);    /* 1 */
    }

    if (off_x || off_y)
	gnome_print_moveto(self->pc, self->state.last_x + off_x,
			   self->state.last_y + off_y);

    /**********************************/
    gnome_print_glyphlist(self->pc, gl);
    /**********************************/

    if (self->state.text_angle != 0)
	nxplot_restorestate(nxplot);

    /*
     * Set the pen position to the end of the text
     */
    if (self->state.text_angle != 0) {
#ifdef HAVE_SINCOS
	sincos(self->state.text_angle*M_PI/180, &s, &c);
#else
	double theta = self->state.text_angle*M_PI/180;
	s = sin(theta);
	c = sin(theta);
#endif
    } else {
	c = 1.0;
	s = 0.0;
    }

    switch (halign)
    {
    case NXPLOT_HALIGN_LEFT:
	last_x += width*c;
	last_y += width*s;
	break;
    case NXPLOT_HALIGN_CENTER:
	last_x += width/2*c;
	last_y += width/2*s;
	break;
    case NXPLOT_HALIGN_RIGHT:
	break;
    default:
	g_warning("Invalid horizontal alignment %i", halign);
    }
    nxplot_move(nxplot, last_x, last_y);

    gnome_glyphlist_unref(gl);
}


static double nxplot_gp_labelwidth(Nxplot *nxplot, const char *text)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    return gnome_font_get_width_utf8(self->font, text);
}


static NxplotGPState* nxplot_gp_state_new(void)
{
    return g_new0(NxplotGPState, 1);
}


static void nxplot_gp_state_free(NxplotGPState *state)
{
    g_return_if_fail(state);
    g_free(state);
}


static void nxplot_gp_state_copy(NxplotGPState *dest, NxplotGPState *src)
{
    g_return_if_fail(dest);
    g_return_if_fail(src);
    memcpy(dest, src, sizeof(NxplotGPState));
}


static NxplotGPState * nxplot_gp_state_dup(NxplotGPState *state)
{
    NxplotGPState *new_state;
    g_return_val_if_fail(state, NULL);
    new_state = nxplot_gp_state_new();
    nxplot_gp_state_copy(new_state, state);
    return new_state;
}


static void nxplot_gp_savestate(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    g_queue_push_tail(self->saved_states, 
		      nxplot_gp_state_dup(&self->state));
    gnome_print_gsave(self->pc);
}

/**
 * nxplot_gp_state_apply:
 * @nxplot: 
 * 
 * Apply the current state settings, i.e. enforce them, updating other
 * NxplotGP fields to reflect the current state.
 **/
static void nxplot_gp_state_apply(Nxplot *nxplot)
{
/*     NxplotGP *self = NXPLOT_GP(nxplot); */
/*     update_wlinewidth(self); */
/*     update_dashes(self); */
/*     nxplot_gp_update_affine(self); */
/*     nxplot_gp_update_font_size(self); */
}


static void nxplot_gp_clear_states(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    NxplotGPState *state = NULL;
    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_endpath(nxplot);
    while (!g_queue_is_empty(self->saved_states))
	state = g_queue_pop_tail(self->saved_states);
    if (state) {
	nxplot_gp_state_copy(&self->state, state);
	nxplot_gp_state_free(state);
	nxplot_gp_state_apply(nxplot);
    }
}


static void nxplot_gp_restorestate(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    NxplotGPState *state;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(!g_queue_is_empty(self->saved_states));
    if (self->inpath) nxplot_endpath(nxplot);
    state = g_queue_pop_tail(self->saved_states);
    nxplot_gp_state_copy(&self->state, state);
    nxplot_gp_state_free(state);
    nxplot_gp_state_apply(nxplot);
    gnome_print_grestore(self->pc);
}




/* static void nxplot_gp_savestate(Nxplot *nxplot) */
/* { */
/*     NxplotGP *self = NXPLOT_GP(nxplot); */

/*     g_return_if_fail(nxplot->opened); */
/*     gnome_print_gsave(self->pc); */
/* } */


/* static void nxplot_gp_restorestate(Nxplot *nxplot) */
/* { */
/*     NxplotGP *self = NXPLOT_GP(nxplot); */

/*     g_return_if_fail(nxplot->opened); */
/*     gnome_print_grestore(self->pc); */
/* } */


static void nxplot_gp_setmatrix(Nxplot *nxplot, double m0, double m1,
				double m2, double m3, double tx, double ty)
{
/*     NxplotGP *self = NXPLOT_GP(nxplot); */

    g_return_if_fail(nxplot->opened);
    /* TODO */
}


static void nxplot_gp_concat(Nxplot *nxplot, double m0, double m1,
			     double m2, double m3, double tx, double ty)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    gdouble matrix[6] = {m0, m1, m2, m3, tx, ty};
    g_return_if_fail(nxplot->opened);
    gnome_print_concat(self->pc, matrix);
    art_affine_multiply(self->state.ctm, self->state.ctm, matrix);
}


static void nxplot_gp_rotate(Nxplot *nxplot, double theta)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    gdouble matrix[6];

    g_return_if_fail(nxplot->opened);
    gnome_print_rotate(self->pc, theta);
    art_affine_rotate(matrix, theta);
    art_affine_multiply(self->state.ctm, self->state.ctm, matrix);
}


static void nxplot_gp_scale(Nxplot *nxplot, double sx, double sy)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    gdouble matrix[6];

    g_return_if_fail(nxplot->opened);
    gnome_print_scale(self->pc, sx, sy);
    art_affine_scale(matrix, sx, sy);
    art_affine_multiply(self->state.ctm, self->state.ctm, matrix);
}


static void nxplot_gp_translate(Nxplot *nxplot, double tx, double ty)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    gdouble matrix[6];

    g_return_if_fail(nxplot->opened);
    gnome_print_translate(self->pc, tx, ty);
    art_affine_translate(matrix, tx, ty);
    art_affine_multiply(self->state.ctm, self->state.ctm, matrix);
}


static double nxplot_gp_descender(Nxplot *nxplot)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    return -gnome_font_get_descender(self->font);
}


static void nxplot_gp_setcliprect(Nxplot *nxplot, NxplotRect *clip_rect)
{
    NxplotGP *self = NXPLOT_GP(nxplot);

    if (self->inpath) nxplot_endpath(nxplot);
    gnome_print_newpath(self->pc);
    gnome_print_moveto(self->pc, clip_rect->x1, clip_rect->y1);
    gnome_print_lineto(self->pc, clip_rect->x2, clip_rect->y1);
    gnome_print_lineto(self->pc, clip_rect->x2, clip_rect->y2);
    gnome_print_lineto(self->pc, clip_rect->x1, clip_rect->y2);
    gnome_print_lineto(self->pc, clip_rect->x1, clip_rect->y1);
    gnome_print_clip(self->pc);
}


static void nxplot_gp_fillcolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    self->state.fillcolor.red   = red;
    self->state.fillcolor.green = green;
    self->state.fillcolor.blue  = blue;
}


static void nxplot_gp_fillcolorname(Nxplot *nxplot, const char *name)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    gint rv;
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    rv = gdk_color_parse(name, &self->state.fillcolor);
    if (!rv) g_warning("Parsing of color '%s' failed", name);
}


static void nxplot_gp_fillopacity(Nxplot *nxplot, guint16 opacity)
{
    NxplotGP *self = NXPLOT_GP(nxplot);
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_gp_endpath(nxplot);
    self->state.fillopacity = opacity / 65535.0;
}

static GnomeGlyphList* nxplot_gp_text_layout(NxplotGP *self, const char *text)
{
    GnomeGlyphList *gl;
    gunichar        chr;
    gint            glyph;

    g_return_val_if_fail(text, NULL);

    gl = gnome_glyphlist_new();
    gnome_glyphlist_advance(gl, TRUE);
    gnome_glyphlist_kerning(gl, 0); /* FIXME: gnomeprint does not do kerning at the
				     * moment. When it does, we need to activate kerning
				     * here (change 0 to 1.0) and enable it also in
				     * nxplot-artft.c */
    gnome_glyphlist_letterspace(gl, 0);
    gnome_glyphlist_font(gl, self->font);
    gnome_glyphlist_color(gl, self->state.foreground|0xff);
    while ((chr = g_utf8_get_char(text)))
    {
	text = g_utf8_next_char(text);
	glyph = gnome_font_lookup_default(self->font, chr);
	gnome_glyphlist_glyph(gl, glyph);
    }
    return gl;
}

static void nxplot_gp_labelmetrics(Nxplot *nxplot, const char *text,
				   NxplotLabelMetrics *metrics)
{
    GnomeGlyphList *gl;
    NxplotGP *self = NXPLOT_GP(nxplot);
    ArtDRect bbox;
    
    gl = nxplot_gp_text_layout(self, text);
    if (!gnome_glyphlist_bbox(gl, NULL /*self->state.ctm*/, 0, &bbox)) {
	metrics->ascender_max  = 0;
	metrics->descender_max = 0;
	metrics->ascender      = 0;
	metrics->descender     = 0;
	metrics->width         = 0;
    } else {
	g_message("labelmetrics: bbox: x0=%f; x1=%f; y0=%f; y1=%f; fontsize=%f",
		  bbox.x0, bbox.x1, bbox.y0, bbox.y1, self->state.font_size);
	metrics->ascender_max  =  gnome_font_get_ascender(self->font);
	metrics->descender_max = -gnome_font_get_descender(self->font);
	metrics->ascender      =  bbox.y1;
	metrics->descender     = -bbox.y0;
	metrics->width         =  bbox.x1 - bbox.x0;
    }
    gnome_glyphlist_unref(gl);
}

